/*
File:    auxiliary.c
Purpose: on-line help and auxiliary functions

dvda-author  - Author a DVD-Audio DVD

(C) Dave Chapman <dave@dchapman.com> 2005
(C) Revised version with zone-to-zone linking Fabrice Nicol <fabnicol@users.sourceforge.net> 2007, 2008

The latest version can be found at http://dvd-audio.sourceforge.net

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>

#include <stdarg.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#include <time.h>

#include "version.h"
#include "winport.h"
#include "audio2.h"
#include "auxiliary.h"
#include "dvda-author.h"
#include "c_utils.h"
#include "file_input_parsing.h"

extern globalData globals;


void version()
{
  foutput("%s%s\n%s", "dvda-author version ", VERSION, "\nCopyright  2005 Dave Chapman; 2007, 2008 Fabrice Nicol\n ");
  foutput("%s","Latest version available from http://dvd-audio.sourceforge.net/\n\n");
  foutput("%s","This is free software; see the source for copying conditions.  There is NO\n");
  foutput("%s","warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\nWritten by Dave Chapman and Fabrice Nicol.\n\n");
}

void help()
{

#ifdef __WIN32__
  system("mode con cols=85 lines=50");
  system("title DVD-A author Help");
#endif


  foutput("%s", "\n\ndvda-author creates high-resolution DVD-Audio discs from .wav and .flac audio files.\n\n");
  foutput("%s","Usage: dvda-author [OPTION]...\n");
  foutput("%s","\nOptions:");
  foutput("%s","\n        \n --help\n");
  foutput("%s","\n        \n -h        Diplays this help.\n");
  foutput("%s","\n        \n --version Diplays version.\n");
  foutput("%s","\n        \n -q        Quiet mode.\n");
  foutput("%s","\n        \n -P        Inserts a final pause before exiting.\n");
  foutput("%s","\n        \n -l        Ouputs a log to filepath.\n");
  foutput("%s","\n        \n -p NNN    specifies the number of the first offset\n           of the AUDIO_PP.IFO file in the output of mkisofs\n");
  foutput("%s","\n        \n -g        You may specify up to 9 groups of tracks. Minimum: one group.\n");
  foutput("%s","\n        \n           Enter full path to files if input directory is not set by [-i].\n");
  foutput("%s","\n        \n -i        Input directory with audio files. For each subdirectory a group\n           will be created. Subdirectory names must comprise\n           one letter followed by one number between 1 and 9\n           possibly followed by letters.");
  foutput("%s","\n        \n -o        Output directory.\n");
  foutput("%s", "\n       \n -V        Path to VIDEO_TS directory");
  foutput("%s","\n        \n -T        Rank of video titleset linked to in video zone (XX in VTS_XX_0.IFO) .");
  foutput("%s","\n        \n           In this case the path to the VIDEO_TS linked to must be indicated.\n");
  foutput("%s","\n        \n -n        Does not generate an empty VIDEO_TS directory.\n");
  foutput("%s","\n        \n -w        Access rights (octal)");
  foutput("%s","\n        \n -d        Increased verbosity (debugging level)");
  foutput("%s","\n        \n -U (-)x   Enter lag to be added/substracted to title length, in 10E-2 second. ");
  foutput("%s","\n\nSupported audio types: .wav .flac\n\n");
  foutput("%s","\nThere must be a maximum of 9 audio groups.\n\n");
  foutput("%s","\nEach subdirectory of an audio input directory will contain titles\nfor a separate audio group.\nA number between 1 and 9 must be included as the second character\nof the subdirectory relative name.\n");
  foutput("%s", "\nExamples:\n");
  foutput("%s", "\n\
          -creates a 3-group DVD-Audio disc (legacy syntax):\n\n\
          dvda-author -g file1.wav file2.flac -g file3.flac -g file4.wav");
  foutput("%s", "\n\n\
          -creates a hybrid DVD disc with both AUDIO_TS mirroring audio_input_directory\n\
          and VIDEO_TS imported from directory VID, outputs disc structure to directory\n\
          DVD_HYBRID and links video titleset #2 of VIDEO_TS to AUDIO_TS:\n\n\
          dvda-author -i ~/audio/audio_input_directory -o DVD_HYBRID -V ~/Video/VID -T 2 \n");
  foutput("%s", "\nReport bugs to fabnicol@users.sourceforge.net\n");
  return;
}



_Bool increment_ngroups_check_ceiling(int *ngroups, void * nvideolinking_groups)
{

  if (*ngroups < 9)
    {
      (*ngroups)++;
      if (nvideolinking_groups != NULL) (*(int *)nvideolinking_groups)++;
    }
  else
    {
      if (nvideolinking_groups != NULL)
        foutput("[ERR]  DVD-Audio only supports up to 9 groups; audio groups=%d; video-linking groups=%d\n", *ngroups, *(int *)nvideolinking_groups);
      else
        foutput("[ERR]  DVD-Audio only supports up to 9 groups; audio groups=%d\n", *ngroups);
      clean_exit(EXIT_SUCCESS, DEFAULT);
    }
  return 1;
}

fileinfo_t** dynamic_memory_allocate(fileinfo_t **  files,int* ntracks,  int  ngroups, int n_g_groups, int nvideolinking_groups)
{
  float memory=0;
  int i, j;
  if ((files= (fileinfo_t **) calloc(ngroups, sizeof(fileinfo_t *))) == NULL)
    EXIT_ON_RUNTIME_ERROR

    for (i=0 ; i < n_g_groups; i++)
      {
        if ((files[i]=(fileinfo_t *) calloc(ntracks[i], sizeof(fileinfo_t))) == NULL)
          EXIT_ON_RUNTIME_ERROR

          memory+=(float) (ntracks[i])*sizeof(fileinfo_t)/1024;

        if (globals.debugging)
          foutput("[MSG]  g-type  audio group  :  %d   Allocating:  %d  track(s)  (strings=%.1f kB)\n", i,  ntracks[i], memory);
      }
  for (i=n_g_groups ; i < ngroups-nvideolinking_groups; i++)
    {

      if ((files[i]=(fileinfo_t *) calloc(ntracks[i], sizeof(fileinfo_t)) )== NULL)
        EXIT_ON_RUNTIME_ERROR

        for (j=0; j < ntracks[i]; j++)
          if ((files[i][j].filename=calloc(CHAR_BUFSIZ, sizeof(char)) )== NULL)
            EXIT_ON_RUNTIME_ERROR

            memory+=(float) (ntracks[i])*(sizeof(fileinfo_t) + CHAR_BUFSIZ)/1024; // CHAR_BUFSIZ characters assigned later on by strdup
      if (globals.debugging)
        foutput("[MSG]  Directory audio group:  %d   Allocating:  %d  track(s)  (strings=%.1f kB)\n", i,  ntracks[i], memory);
    }
  for (i=ngroups-nvideolinking_groups ; i < ngroups; i++)
    {
      if ((files[i]=(fileinfo_t *) calloc(1, sizeof(fileinfo_t))) == NULL)
        EXIT_ON_RUNTIME_ERROR
        memory+=(float) sizeof(fileinfo_t)/1024;
      /* sanity check: 0 tracks should be allocated */
      if (globals.debugging)
        foutput("[MSG]  Video-linking group  :  %d   Allocating:  %d  track(s)  (strings=%.1f kB)\n", i, ntracks[i], memory);
    }

  return files;
}

