/*

dvda-author2.c  - Author a DVD-Audio DVD

Copyright Fabrice Nicol <fabnicol@users.sourceforge.net>

The latest version can be found at http://dvd-audio.sourceforge.net

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>

#include <dirent.h>
#include <locale.h>
#include <errno.h>
#include <inttypes.h>


#include "c_utils.h"
#include "audio2.h"
#include "ats.h"
#include "atsi.h"
#include "amg.h"
#include "samg.h"
#include "auxiliary.h"
#include "commonvars.h"
#include "videoimport.h"
#include "dvda-author.h"
#include "winport.h"
#include "file_input_parsing.h"


/*  Global  options */

globalData globals;

#ifndef __WIN32__
struct rusage nothing, start, end;
#endif


int main(int argc,  char* const argv[])
{

  extern char *optarg;
  extern int optind;
  const char* allowed_options="ghi:l:T:no:p:PqV:U:w:d-:";

  _Bool   copy_VIDEO_TS=0;

  int i,j, k, c;
  int error, errmsg;
  int ngroups=0, nvideolinking_groups=0;
  int ntracks[9];
  int atsi_sectors[9];
  int32_t startsector=(int32_t) STARTSECTOR, last_sector;

  uint64_t totalsize=0;
  uint64_t  sector_pointer_VIDEO_TS=0;
  int VTSI_rank[MAXIMUM_LINKED_VTS];
  int maximum_VTSI_rank=1;
  int access_rights=0777;
  int Udo=0;
  DIR *dir;
  parse_t  audiodir;

  /*  DEFAULT   SETTINGS  */
  // Locale, time and log management

  setlocale(LC_ALL, "en_US");
  starter();

  // Settings are hard-code set by default as follows:


  globals.settings.logfile=LOGFILE;
  globals.settings.indir=INDIR;
  globals.settings.outdir=OUTDIR;
  globals.settings.linkdir=LINKDIR;

  /* by default, globals booleans are set as nul*/
  /*  command line parsing is silent by default*/

  //globals.videolinking=0;
  //globals.end_pause=0;
  //globals.logfile=0;
  //globals.videozone=0;
  //globals.silence=0;
  //globals.debugging=0;


  // If a default setting textfile exists, it overrides hard-code defaults
  // Path to defaults settings file must be "Path/to/DVD-A author folder/defaults.txt"
  // Default settings, either hard-code set or by defaults textfile, can be overridden by command-line.

// TO DO: read_defaults();
  errno=0;

  /*  INITIAL VALUES  */

  // Global data are initialised in the header.

  memset(ntracks, 0, sizeof(ntracks));

  // management of video-linking groups uses default 0-padding throughout; it is necessary to ensure null initialization of files[][]

  ngroups=0;
  nvideolinking_groups=0;

  /* getopt is now used for command line parsing. To ensure compatibility with prior "Dave" versions, the easier way out
  *  is to duplicate the command line. Otherwise getopt reorders options/non-options and multiple arguments of -g ...
  *  are consequently misplaced */
  char  *argv_scan[argc];

  for (k=0; k<argc; k++)
    {
      if ((argv_scan[k]=strdup(argv[k])) == NULL)
        EXIT_ON_RUNTIME_ERROR
      }

  /* COMMAND-LINE  PARSING: first pass for global behaviour option: log, help, version, verbosity */

  while ((c=getopt(argc, argv_scan, allowed_options)) != -1)
    {
      switch (c)
        {

        case '-' :
          if  (strcmp(optarg, "version") == 0)
            {
              globals.silence=0;
              version();
              if (globals.end_pause) pause_dos_type();
              clean_exit(EXIT_SUCCESS, DEFAULT);
              break;
            }

          if  (strcmp(optarg, "help") == 0)
            /* same as below */

          case 'h' :
          help();
          globals.silence=0;
          if (globals.end_pause) pause_dos_type();
          clean_exit(EXIT_SUCCESS, DEFAULT);
          break;

        case 'q' :
          globals.silence=1;
          break;

        case 'd':
          globals.debugging=1;
          globals.silence=0; // overrides -q;
          break;

        case 'l' :

          globals.settings.logfile=optarg;
          if ((globals.journal=fopen(globals.settings.logfile, "a")) == NULL)
            fprintf(stderr, "[ERR]  Logfile %s could not be opened", globals.settings.logfile);
          globals.logfile=1;

          break;

        }
    }

  if (globals.logfile) foutput("%s%s\n", "[PAR]  Log file is: ", globals.settings.logfile);

  INITIALISE


  /* COMMAND-LINE PARSING: second pass to determine memory allocation (thereby avoiding heap loss)
   * We give up getopt here to allow for legacy "Dave" syntax with multiple tracks as -g arguments
   * (not compatible with getopt or heavy to implement with it.  */

  if (globals.debugging) foutput("\n%s\n", "[INF]  First scan of track list for memory allocation...");


  for (k=0; k < argc; k++)
    {
      if (argv[k][0] != '-') continue;
      if (argv[k][1] == 'g')
        {
          k++;
          for (; (k < argc)&&(argv[k][0] !='-'); k++)
            ntracks[ngroups]++;

          increment_ngroups_check_ceiling(&ngroups, NULL);
          k--;
        }
    }


  int n_g_groups=ngroups;
  /* crucial */
  optind=0;
  fileinfo_t **files_dummy=NULL;

  /* COMMAND-LINE PARSING: third pass  to determine memory allocation with non-g options and getopt */

  while ((c=getopt(argc, argv_scan, allowed_options)) != -1)
    {
      switch (c)
        {

        case 'i' :

          globals.settings.indir=strdup(optarg);

          foutput("%s%s\n", "[PAR]  Input directory is: ", 	optarg);
          DIR *dir;

          if ((dir=opendir(optarg)) == NULL)
            EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Input directory could not be opened")

            change_directory(globals.settings.indir);

          audiodir=parse_directory(dir, ntracks, n_g_groups, 0, files_dummy);

          ngroups=audiodir.ngroups;

          memmove(ntracks, audiodir.ntracks, 9*sizeof(int));

          if (closedir(dir) == -1)
            foutput( "%s\n", "[ERR]  Impossible to close dir");

          break;

        case 'T':

          if (nvideolinking_groups == MAXIMUM_LINKED_VTS)
            {
              foutput("[ERR]  Error: there must be a maximum of %d video linking groups\n      Ignoring additional links...\n\n", MAXIMUM_LINKED_VTS);
              break;
            }

          // VTSI_rank is the rank of the VTS that is linked to in VIDEO_TS directory
          VTSI_rank[nvideolinking_groups]=atoi(optarg);

          if   (VTSI_rank[nvideolinking_groups] > 99)
            EXIT_ON_RUNTIME_ERROR_VERBOSE( "[ERR]  There must be a maximum of 99 video titlesets in video zone. Try again...\n\n")

            if (nvideolinking_groups == 0)
              maximum_VTSI_rank=VTSI_rank[nvideolinking_groups];
            else
              maximum_VTSI_rank=MAX(VTSI_rank[nvideolinking_groups], maximum_VTSI_rank);

          globals.videozone=1;
          globals.videolinking=1;

          increment_ngroups_check_ceiling(&ngroups, &nvideolinking_groups);

          break;
        }
    }

  /* Here the group parameters are known: ngroups (total),  n_g_groups (legacy -g syntax), nvideolinking_groups */

  /* command line copy is now useless: freeing space */

  for (k=0; k<argc; k++)
    free(argv_scan[k]);

  /* Performing memory allocation (calloc)
  *
  *  Ordering
  *  -----------
  *  Groups are ordered according to the following order : g-type groups < directory groups < video-linking groups
  *
  *  Allocation
  *  ------------
  *  g-type groups are granted "for free" as they are allocated by command-line argv parsing
  *  Directory groups are costly as they must be allocated freshly
  *  Video-linking groups have just one track and are reordered with highest ranks */

  fileinfo_t ** files=NULL;

  files=dynamic_memory_allocate(files, ntracks, ngroups, n_g_groups, nvideolinking_groups);

  /* COMMAND-LINE PARSING: fourth pass to assign filenames without allocating new memory (pointing to argv) */

  int m, ngroups_scan=0;

  if ((n_g_groups)&&(globals.debugging)) foutput("%s", "[INF]  Assigning g-type filenames...\n");

  for (k=0; k < argc; k++)
    {
      if (argv[k][0] != '-') continue;
      if (argv[k][1] == 'g')
        {
          k++;

          for (m=0; (m+k < argc)&&(argv[m+k][0] !='-'); m++)
            {
              if (globals.debugging) foutput("       files[%d][%d].filename=%s\n", ngroups_scan, m, argv[m+k]);
              files[ngroups_scan][m].filename=argv[m+k];

            }
          k+=m-1;
          ngroups_scan++;

        }
    }


  foutput("%s", "\n\n");

  /* COMMAND-LINE  PARSING: fourth pass for main arguments and non-g filename assignment */


  // Changing scanning variable names for ngroups_scan and nvideolinking_groups_scan

  // n_g_groups=ngroups_scan; (useless)

  ngroups_scan=0;
  int nvideolinking_groups_scan=0;

  /* crucial */
  optind=0;

  while ((c=getopt(argc, argv, allowed_options)) != -1)
    {

      switch (c)
        {


        case 'T':

          ngroups_scan++;
          nvideolinking_groups_scan++;

          // allowing for a single title in video-linking group
          //  videolinkg groups are allocated in last position whatever the form of the command line
          ntracks[ngroups-nvideolinking_groups+nvideolinking_groups_scan-1]=0;

          files[ngroups-nvideolinking_groups+nvideolinking_groups_scan-1][0].first_PTS=0x249;
          // all other characteristics of videolinking titles ar null (handled by memset)

          break;

        case 'V' :
          //  video-linking directory to VIDEO_TS structure
          globals.videolinking=1;
          globals.videozone=1;
          globals.settings.linkdir=optarg;
          foutput("%s%s\n", "[PAR]  VIDEO_TS input directory is: ", optarg);
          copy_VIDEO_TS=1;

          break;

        case 'n' :
          // There is no videozone in this case

          if (globals.videolinking)
            {
              foutput("%s\n", "[WAR]  You cannot delete video zone with -n if -V is activated too.\n      Ignoring -n...");
              break;
            }
          globals.videozone=0;
          foutput("%s\n", "[PAR]  No video zone");
          break;

        case 'i' :

          if ((dir=opendir(optarg)) == NULL)
            EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Input directory could not be opened")

            change_directory(globals.settings.indir);

          parse_directory(dir, ntracks, n_g_groups, READTRACKS, files);

          if (closedir(dir) == -1)
            foutput( "%s\n", "[ERR]  Impossible to close dir");


          /* all-important, otherwise irrelevant EXIT_ON_RUNTIME_ERROR will be generated*/

          errno=0;

          break;

        case 'o' :
          globals.settings.outdir=optarg;
          foutput("%s%s\n", "[PAR]  Output directory is: ", optarg);
          break;

        case 'p' :
          startsector=(int32_t) strtol(optarg, NULL, 0)/2048;
          errmsg=errno;
          switch (errmsg)
            {
            case   EINVAL :
              foutput( "%s\n",  "[ERR]  Incorrect offset value");
              clean_exit(EXIT_SUCCESS, DEFAULT);
              break;
            case   ERANGE :
              EXIT_ON_RUNTIME_ERROR_VERBOSE( "[ERR]  Offset range--overflowing LONG INT.");
              break;
            }
          errno=0;
          foutput("[MSG]  Using starting sector of %"PRId32"\n", startsector);
          break;

        case 'P':
          globals.end_pause=1;
          foutput("%s\n", "[PAR]  No end pause");
          break;

        case 'w':
          /* input must be in octal form */
          access_rights=(mode_t) strtol(optarg, NULL, 8);
          foutput("[PAR]  Access rights=%d\n", access_rights);
          break;

        case 'U':

          /* Udo switch: -U x, where x is expressed in 10E-2 second;
          * e.g: -U -60 takes off 60/100 second out of PTS ticks; -U 60 adds it. */

          Udo=atoi(optarg)*900;  // expressed in PTS ticks
          foutput("[PAR]  Special Udo switch=%+d PTS ticks\n", Udo);
          break;

        }

    }


  errno=0;
  char audiotsdir[CHAR_BUFSIZ+9];
  char videotsdir[CHAR_BUFSIZ+9];
  int naudio_groups=ngroups-nvideolinking_groups;

  make_directory(globals.settings.outdir, access_rights, DEFAULT);

  if ((audiotsdir == NULL) || (videotsdir == NULL))
    EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Could not allocate filepaths to _TS directory");

  STRING_WRITE_CHAR_BUFSIZ(audiotsdir, "%s/AUDIO_TS", globals.settings.outdir)

  make_directory(audiotsdir, access_rights, DEFAULT);
  errno=0;

  if (globals.videozone)
    {
      STRING_WRITE_CHAR_BUFSIZ(videotsdir, "%s/VIDEO_TS", globals.settings.outdir)
      if (! globals.videolinking)  make_directory(videotsdir, access_rights, DEFAULT);
      errno=0;
    }


  /* Step 1 - parse all audio files and store the file formats, lengths etc */

  foutput("\n\n%s\n", "DVD Layout:\n\n");
  foutput("%s\n", "Group  Track    Rate Bits  Ch       Length  Filename\n");

  for (i=0; i < naudio_groups;i++)
    {
      for (j=0; j < ntracks[i];  j++)
        {
          error=wav_getinfo(&files[i][j]);

          if (error)
            {
              error=flac_getinfo(&files[i][j]);
            }

          files[i][j].PTS_length += Udo;

          if (error)
            {
              foutput("[ERR]  Cannot open %s\n",files[i][j].filename);
              EXIT_ON_RUNTIME_ERROR
            }
          foutput("    %d     %02d  %6"PRIu32"   %02d   %d   %10"PRIu64"   ",i+1, j+1, files[i][j].samplerate, files[i][j].bitspersample, files[i][j].channels, files[i][j].numsamples);
          foutput("%s\n",files[i][j].filename);
          totalsize+=files[i][j].numbytes;
        }
    }
  foutput("%c\n", '\n');

  foutput("[MSG]  Size of raw PCM data: %"PRIu64" bytes (%.2f  MB)\n\n",totalsize, (float) totalsize/(1024*1024));

  /* Audio zone system file  parameters  */

  for (i=0; i < naudio_groups; i++)
    {
      error=create_ats(audiotsdir,i+1,&files[i][0],ntracks[i]);
      error=create_atsi(audiotsdir,i+1,&files[i][0],ntracks[i],&atsi_sectors[i]);
    }

  /* Creating AUDIO_PP.IFO */

  last_sector=create_samg(
                audiotsdir,
                files,
                ntracks,
                ngroups,
                nvideolinking_groups,
                atsi_sectors);

  /*   sector_pointer_VIDEO_TS= number of sectors for AOBs + 2* SIZE_AMG + 2* SIZE_ATS*ngroups   */

  sector_pointer_VIDEO_TS= 2*SIZE_AMG;

  for (i=0; i < naudio_groups; i++)
    {
      for (j=0; j < ntracks[i]; j++)
        {
          sector_pointer_VIDEO_TS+=files[i][j].last_sector - files[i][j].first_sector+1;
        }
      sector_pointer_VIDEO_TS +=2*atsi_sectors[i];
    }

  if (globals.debugging)
    {
      foutput( "       Sector pointer to VIDEO_TS from AUDIO_TS= %"PRIu64" sectors\n", sector_pointer_VIDEO_TS);
      foutput( "\n%s", "[INF]  Checking coherence of pointers...");

      if (SIZE_SAMG+STARTSECTOR+sector_pointer_VIDEO_TS != last_sector+1+atsi_sectors[naudio_groups-1])

        foutput("\n[WAR]  Pointers to VIDEO_TS are not coherent %"PRIu64" , %"PRIu32"\n",
                SIZE_SAMG+STARTSECTOR+sector_pointer_VIDEO_TS, (uint32_t) last_sector+1+atsi_sectors[naudio_groups-1]);
      else
        foutput("%s\n", " OK");
    }

  foutput( "\n[MSG]  Total size of AUDIO_TS: %"PRIu64" sectors\n", sector_pointer_VIDEO_TS + SIZE_SAMG);

  foutput( "[MSG]  Start offset of  VIDEO_TS in ISO file: %"PRIu64" sectors,  offset %"PRIu64"\n\n", sector_pointer_VIDEO_TS + SIZE_SAMG + startsector,
           (sector_pointer_VIDEO_TS + SIZE_SAMG + startsector)*2048);

  /* Creating AUDIO_TS.IFO */

  uint32_t  relative_sector_pointer_VTSI[nvideolinking_groups];
  uint32_t  videotitlelength[nvideolinking_groups];

  memset(relative_sector_pointer_VTSI, 0, nvideolinking_groups*4);
  memset(videotitlelength, 0, nvideolinking_groups*4);

//  relative_sector_pointer_VTSI=absolute_se+relative_sector_pointer_in_VTS

  /*
  *   Version 200806: added to function create_amg:
  *
  *   int VTSI_rank[N]
  *								VTSI_rank[k]	= rang of k-th video titleset linked to in video zone (< 10)
  *
  *   uint32_t  relative_sector_pointer_VTSI[N]
  *								relative_sector_pointer_VTSI[k] = & VTS_XX_0.IFO -&AUDIO_TS.IFO, expressed in sectors, in which XX=VTSI_rank[k]
  *
  *   uint32_t  videotitlelength[[N]
  *							   videotitlelength[k] = length of title linked to in PTS ticks
  *
  *   N= number of video linking groups in audio zone ( + number of audio groups < 10)
  *
  *
  */

// returns relative_sector_pointer_VTSI and videotitlelength

  if (globals.videolinking)
    {

      get_video_system_file_size(globals.settings.linkdir, maximum_VTSI_rank, sector_pointer_VIDEO_TS, relative_sector_pointer_VTSI);
      get_video_PTS_ticks(globals.settings.linkdir, videotitlelength, nvideolinking_groups, VTSI_rank);

      char    newpath[CHAR_BUFSIZ];
      STRING_WRITE_CHAR_BUFSIZ(newpath, "%s%s", globals.settings.outdir, "/VIDEO_TS")
      if (copy_VIDEO_TS)
        copy_directory(globals.settings.linkdir, newpath, access_rights);
    }

  create_amg(
    audiotsdir,
    files,
    ntracks,
    ngroups,
    nvideolinking_groups,
    atsi_sectors,
    videotitlelength,
    VTSI_rank,
    relative_sector_pointer_VTSI);

  EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Audio zone system files could not be created.")

  errno=0;

  foutput("%c\n", '\n');
  foutput("%s\n" , "Group  Track    First Sect   Last Sect  First PTS   PTS length\n");

  for (i=0; i < ngroups; i++)
    {
      for (j=0; j < ntracks[i]; j++)
        {
          foutput("    %d     %02d  %10"PRIu32"  %10"PRIu32"  %10"PRIu64"  ",i+1,j+1,files[i][j].first_sector,files[i][j].last_sector,files[i][j].first_PTS);
          foutput("%10"PRIu64"\n",files[i][j].PTS_length);
        }
    }

  SINGLE_DOTS
  COMPUTE_EXECTIME
  SINGLE_DOTS

  /* freeing */

  for (i=n_g_groups; i < ngroups-nvideolinking_groups; i++)
    for (j=0; j < ntracks[i]; j++)
      {
        free(files[i][j].filename);
        if (globals.debugging)
          foutput("[INF]  Freeing i=%d  j=%d\n",i, j );
      }

  for (i=0; i < ngroups; i++)
    free(files[i]);
  free(files);

  if (globals.end_pause) pause_dos_type();
  return(errno);

}

