/*
File:    file_input_parsing.c
Purpose: parses input directories

dvda-author  - Author a DVD-Audio DVD

Copyright Fabrice Nicol <fabnicol@users.sourceforge.net> 2007, 2008

The latest version can be found at http://dvd-audio.sourceforge.net

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include "file_input_parsing.h"

extern globalData globals;

int read_tracks(char  *full_path, int *ntracks, char * parent_directory, char* filename, int ngroups_scan)
{

  if (ntracks[ngroups_scan] < 99)
    {

      if (parent_directory != NULL)
        {

          snprintf(full_path, CHAR_BUFSIZ, "%s%c%s%c%s", globals.settings.indir, '/', parent_directory,'/', filename);
        }
      else
        STRING_WRITE_CHAR_BUFSIZ(full_path, "%s", filename)

      }
  else
    {
      foutput("[MSG] Error: Too many input files specified - group %d, track %d\n",ngroups_scan,ntracks[ngroups_scan]);
      clean_exit(EXIT_SUCCESS, DEFAULT);
    }

  return(ntracks[ngroups_scan]);
}


parse_t
parse_directory(DIR* dir,  int* ntracks, int n_g_groups, int action, fileinfo_t **files)
{

  int ngroups_scan=0, ngroups=0, control=0;
  struct dirent *rootdirent;

  parse_t audiodir;

  if (globals.debugging) foutput("%s\n", "[INF]  Parsing audio input directory");

  while ((rootdirent=readdir(dir) )!= NULL)
    {
      if (rootdirent->d_name[0] == '.') continue;

      /* ngroups_scan is the rank of the parsed group read from the user's index in the directory name */
      
      ngroups_scan= rootdirent->d_name[1]-'0';

      if ((ngroups_scan > 10) || (ngroups_scan < 1))
        EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Subdirectories must be labelled ljm, with l, m any letters and j a number between 1 and 10")

        change_directory(rootdirent->d_name);

      DIR *subdir;
      if ((subdir=opendir(".")) == NULL)
        EXIT_ON_RUNTIME_ERROR_VERBOSE("[ERR]  Input directory could not be opened")


        struct dirent * subdirent;
      control++;

      /* Memory allocation will be based on user input ngroups = Max {ngroups_scan}  */

      ngroups=MAX(ngroups_scan, ngroups);

      ntracks[n_g_groups+ngroups_scan-1]=0;

      do
        {

          subdirent=readdir(subdir);
          if (subdirent == NULL)
            {
              closedir(subdir);
              change_directory(globals.settings.indir);
              break;
            }

          if (subdirent->d_name[0] == '.') continue;

          /* assignment of values is based on ngroups_scan */

          ntracks[n_g_groups+ngroups_scan-1]++;

          if (action == READTRACKS)
            {
              char buf[CHAR_BUFSIZ];
              read_tracks(buf, ntracks, rootdirent->d_name, subdirent->d_name, ngroups_scan);

              if (globals.debugging)
                foutput("[INF]  Copying directory files[%d][%d]\n", n_g_groups+ngroups_scan-1, ntracks[n_g_groups+ngroups_scan-1]-1);

              memmove(files[n_g_groups+ngroups_scan-1][ntracks[n_g_groups+ngroups_scan-1]-1].filename, buf, CHAR_BUFSIZ);
            }
          /* readdir may well not scan the subdirectories in the 'right' order even if labelled g1, g2, g3, ....succesively. */

        }
      while (1);
    }

  if (action != READTRACKS) foutput("[MSG]  %d groups/subdirectories were parsed; ngroups=%d\n", control, ngroups);

  /* Controlling for contiguousness of ngroups_scan values; a crash may occur if not ensured; letting it go however */

  if (ngroups != control) foutput("%s", "[WAR]  Critical -- Groups are not labelled contiguously (g1, ... ,gn).\n");

  audiodir.ngroups=ngroups+n_g_groups;
  audiodir.ntracks=ntracks;

  return audiodir;
}

