/*

File:    audio2.h
Purpose: Deal with reading the input audio files.

dvda-author  - Author a DVD-Audio DVD

(C) Dave Chapman <dave@dchapman.com> 2005, revised by Fabrice Nicol, 2008

The latest version can be found at http://dvd-audio.sourceforge.net

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#ifndef _AUDIO_H
#define _AUDIO_H

#include <stdio.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "stream_decoder.h"
#include "dvda-author.h"
#define AFMT_WAVE 1
#define AFMT_FLAC 2
#define N_FILESIN_VIDEOTS 29

typedef struct
  {
    FILE* fp;
    FLAC__StreamDecoder* flac;
    // Used for FLAC decoding:
    uint8_t buf[1024*256];
    int n;
    int bytesread;
    int eos;
  } audio_input_t;

typedef struct
  {
    char *filename;
    int type;
    int bitspersample;
    int sampleunitsize;
    int channels;
    uint32_t samplerate;
    uint64_t numsamples;
    uint64_t numbytes; // theoretical file size
    uint64_t file_size; // file size on disc
    uint64_t bytesperframe;
    uint64_t bytespersecond;
    uint32_t first_sector;
    uint32_t last_sector;
    uint64_t first_PTS;
    uint64_t PTS_length;
    int audio_format; // Pointer to the audio format record
    audio_input_t* audio;  // Used whilst decoding.
  } fileinfo_t;



int wav_getinfo(fileinfo_t* info);
int flac_getinfo(fileinfo_t* info);
void uint32_copy(uint8_t* buf, uint32_t x);
void uint16_copy(uint8_t* buf, uint16_t x);
int audio_open(fileinfo_t* info);
int audio_read(fileinfo_t* info, uint8_t* buf, int count);
int audio_close(fileinfo_t* info);


void read_defaults();




#endif
